#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "CUnit.h"
#include "TestDB.h"
#include "Basic.h"
#include "fat.h"
#include "format.h"
#include "load_fat.h"

/**
 * \file 
 *
 * Questo file definisce la test suite per la funzione di formattazione
 * del filesystem SFAT.
 */
/** Funzione di inizializzazione della suite */
int success_init(void) { return 0; }
/** Funzione di clean up della suite */
int success_clean(void) { return 0; }

/**
 * Questa funzione crea e formatta un filesystem utilizzando i parametri
 * specificati in ingresso. Il test effettua i controlli
 * necessari a verificare che il file sia stato formattato correttamente.
 *
 * \param fname nome del file da formattare.
 * \param bsize dimensione dei blocchi del filesystem
 * \param nblock numero di blocchi disponibili. 
 */
void run_formatter (char *fname, int bsize, int nblock) {
    int ret, len, i, nbytes;
    struct stat st;
    FILE *fs;
    struct fat_ctrl f_ctrl;
    struct boot_sector b_sec;
    struct dir_entry d_entry;
    unsigned int *fat_check;
    char blk_elem;

    /*
     * Creazione del file device che conterra' il file system. 
     * Se il file esiste viene troncato 
     * Se fopen ritorna NULL il test viene abortito 
     */
    fs = fopen(fname, "w");
    CU_ASSERT_PTR_NOT_NULL_FATAL(fs); 

    /*
     * Formattazione con nblock blocchi di 
     * size bsize
     */
    ret = fat_format(fs, bsize, nblock);
    CU_ASSERT_EQUAL(0, ret);
    fclose(fs);

    /*
     * La lunghezza in byte e' giusta?
     */
    len = sizeof(struct boot_sector);
    len = len + sizeof(unsigned int) * nblock;
    len = len + bsize * nblock;
    stat(fname, &st);
    CU_TEST_FATAL(len == st.st_size);

    /*
     * Effettua il mounting del file system e 
     * controlla che la format abbia scritto i valori giusti
     */
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(0, ret);

    /*
     * Controlla i valori del boot sector
     */
    b_sec = f_ctrl.b_sector;
    CU_TEST_FATAL(b_sec.fs_type == FAT_TYPE);
    CU_TEST_FATAL(b_sec.block_size  == bsize);
    CU_TEST_FATAL(b_sec.num_block  == nblock);

    /*
     * Controlla il valore di blk_base
     */
    CU_ASSERT_EQUAL_FATAL(f_ctrl.blk_base, sizeof(struct boot_sector) + (sizeof(unsigned int) * b_sec.num_block));

    /*
     * Controlla i valori nella File Allocation Table.
     * L'array deve contenere BLOCK_FREE in tutte le posizioni 
     * eccetto la prima, che deve contenere LAST_BLOCK.
     */
    fat_check = f_ctrl.fat_ptr;
    for (i=0; i<b_sec.num_block; i++)
      CU_ASSERT_EQUAL_FATAL(fat_check[i], i==ROOT_IDX ? LAST_BLOCK : BLOCK_FREE);

    /*
     * Controllo intermedio.
     * Tutti i blocchi prima di ROOT_IDX devono essere
     * inizializzati a INIT_BLOCK.
     */
    fseek(fs, f_ctrl.blk_base, SEEK_SET);
    nbytes = ROOT_IDX*b_sec.block_size;
    for (i=0; i<nbytes; i++) {
	ret = fread(&blk_elem, sizeof(char), 1, fs);
	CU_ASSERT_EQUAL_FATAL(1, ret);
	CU_ASSERT_EQUAL_FATAL(blk_elem, INIT_BLOCK);
    }

    /*
     * Controlla che la directory root contenga i valori giusti per 
     * "." e ".."
     */
    ret = fread(&d_entry, sizeof(d_entry), 1, fs);
    CU_ASSERT_EQUAL_FATAL(1, ret);
    CU_ASSERT_EQUAL_FATAL(d_entry.used, DIR_ENTRY_BUSY);
    CU_ASSERT_STRING_EQUAL_FATAL(d_entry.name, ".");
    CU_ASSERT_EQUAL_FATAL(d_entry.attr, SUB_ENTRY);
    CU_ASSERT_EQUAL_FATAL(d_entry.index, 0);
    CU_ASSERT_EQUAL_FATAL(d_entry.len, 2 * sizeof(d_entry));

    ret = fread(&d_entry, sizeof(d_entry), 1, fs);
    CU_ASSERT_EQUAL_FATAL(1, ret);
    CU_ASSERT_EQUAL_FATAL(d_entry.used, DIR_ENTRY_BUSY);
    CU_ASSERT_STRING_EQUAL_FATAL(d_entry.name, "..");
    CU_ASSERT_EQUAL_FATAL(d_entry.attr, SUB_ENTRY);
    CU_ASSERT_EQUAL_FATAL(d_entry.index, 0);
    CU_ASSERT_EQUAL_FATAL(d_entry.len, 2 * sizeof(d_entry));

    /*
     * Controllo finale.
     * Tutti i blocchi rimanenti devono essere inizializzati a 
     * INIT_BLOCK.
     */
    len = 2 * sizeof(d_entry) + nbytes;
    nbytes = b_sec.num_block*b_sec.block_size;
    for (i=len; i<nbytes; i++) {
	ret = fread(&blk_elem, sizeof(char), 1, fs);
	CU_ASSERT_EQUAL_FATAL(1, ret);
	CU_ASSERT_EQUAL_FATAL(blk_elem, INIT_BLOCK);
    }  
}

/**
 * \test
 *
 * Testa la formattazione di un filesystem che utilizza 100 blocchi ciascuno
 * di dimensione 128 bytes.
 */
void format_128 (void)
{
    run_formatter("./DataFile/FAT_128", 128, 100);
}

/**
 * \test
 *
 * Testa la formattazione di un filesystem che utilizza 10 blocchi ciascuno
 * di dimensione 1KB.
 */
void format_1K (void)
{
    run_formatter("./DataFile/FAT_1K", 1024, 10);
}

/**
 * \test
 *
 * Testa la formattazione di un filesystem che utilizza 10 blocchi ciascuno
 * di dimensione 2KB bytes.
 */
void format_2K (void)
{
    run_formatter("./DataFile/FAT_2K", 1024*2, 10);
}

/**
 * \test
 *
 * Testa la formattazione di un filesystem che utilizza 10 blocchi ciascuno
 * di dimensione 4KB.
 */
void format_4K (void)
{
    run_formatter("./DataFile/FAT_4K", 1024*4, 10);
}

/** Definizione della suite. */
void add_suite_format(void)
{
    CU_pSuite format_suite = NULL;

    format_suite = CU_add_suite("Format Operation", success_init, success_clean);
    CU_add_test(format_suite, "FAT format/mount Test (128 Bytes)", format_128);
    CU_add_test(format_suite, "FAT format/mount Test (1K Bytes)", format_1K);
    CU_add_test(format_suite, "FAT format/mount Test (2K Bytes)", format_2K);
    CU_add_test(format_suite, "FAT format/mount Test (4K Bytes)", format_4K);
}

