#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <mcheck.h>
#include "CUnit.h"
#include "TestDB.h"
#include "Basic.h"
#include "fat.h"
#include "format.h"
#include "load_fat.h"
#include "list.h"
#include "make_dir.h"
#include "mkdir_suite.h"

/**
 * \file
 * Questo file definisce la test_suite per la funzione di creazione di una 
 * directory \c fat_mkdir e per la funzione di listing di una directory
 * \c fat_ls.
 */


/** Funzione di inizializzazione della suite  */
int mkdir_init(void) {  return 0; }

/** Funzione di inizializzazione della suite  */
int mkdir_clean(void) {    return 0; }

/**
 * \test 
 *
 * Questo test ha il compito di controllare che pathname non corretti vengano
 * riconosciuti. In particolare si controlla che venga generato un errore se
 * si cerca di generare una nuova directory utilizzando i seguenti path:
 *
 * - Path name non assoluti, come ad esempio \c not/abs/path
 * - Path name che usano nomi troppo lunghi (che eccedono gli 8 caratteri).
 */
void test_pathname (void)
{
    FILE *fs;
    struct fat_ctrl f_ctrl;
    int ret;
     
    fs = fopen("./DataFile/FAT_MKDIR", "w");
    fat_format(fs, 128, 10);
    fclose(fs);
    fs = fopen("./DataFile/FAT_MKDIR", "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(0, ret);

    /*
     * Controlla che i path non assoluti vengano scartati
     */
    ret = fat_mkdir(fs, &f_ctrl, NULL);
    CU_ASSERT_EQUAL_FATAL(ret, EBDP);
    ret = fat_mkdir(fs, &f_ctrl, "not/abs/path");
    CU_ASSERT_EQUAL_FATAL(ret, EBDP);

    /*
     * Controlla che i nomi non eccadano MAX_LEN_NAME
     */
    ret = fat_mkdir(fs, &f_ctrl, "/toooooooo/long");
    CU_ASSERT_EQUAL_FATAL(ret, ENTL);
    ret = fat_mkdir(fs, &f_ctrl, "/looooooooooooong");
    CU_ASSERT_EQUAL_FATAL(ret, ENTL);
   
    /*
     * Cleanup
     */
    fclose(fs);
    unlink("./DataFile/FAT_MKDIR"); 
}

/**
 * \test
 *
 * Esegue il test di creazione di una directory. In particolare vengono
 * eseguite le seguenti operazioni:
 *
 * - Formattazione di un filesystem SFAT con \a num_block blocchi
 *   di dimensione \a block_size.
 * - Creazione della sirectory \c /dir1.
 * - Controllo che le informazioni del boot sector siano corrette 
 *   (in particolare non siano state modificate).
 * - Controllo che la directory table della root (/) contenga
 *   le 3 entries ".", "..", "dir1"
 * - Controllo che la directory table per \c dir1 sia stata correttamente
 *   creata.
 * - Controllo che la FAT contenga le corrette informazioni per ciascun
 *   blocco.
 * - Infine controllo che sia possibile effettuare il listing della nuova 
 *   directory.
 *
 * \param fname il nome del file utilizzato per memorizzare il filesystem FAT.
 * \param block_size la dimensione di ciascun blocco del filesystem FAT.
 * \param num_block numero di blocchi utilizzati dal filesystem FAT.
 *
 */
void run_mkdir (char *fname, int block_size, unsigned int num_block)
{
    FILE *fs;
    struct fat_ctrl f_ctrl;
    struct dir_entry root_info;
    unsigned int len, *f_ptr;
    int i, cnt, n_entr, b_size, ret, index_dir1;
    struct stat st;
    char *list;
    char ex_list[ ] = ".\t..\tdir1";
     
    /*
     * Formata il file e ....
     */
    fs = fopen(fname, "w");
    fat_format(fs, block_size, num_block);
    fclose(fs);
    /* ... lo carica !!!! */
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     *  Prova a creare una sottodirectory e chiude il filesystem
     */
    ret = fat_mkdir(fs, &f_ctrl, "/dir1");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    fclose(fs);
    /* 
     * ... prova a ricaricarlo ...
     **/
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Prima controlla il boot sector e poi ...
     */
    CU_ASSERT_EQUAL_FATAL(f_ctrl.b_sector.fs_type, FAT_TYPE); 
    CU_ASSERT_EQUAL_FATAL(f_ctrl.b_sector.block_size, block_size);
    CU_ASSERT_EQUAL_FATAL(f_ctrl.b_sector.num_block, num_block);

    /* ... controlla l'inizio dei dati. */
    len = sizeof(struct boot_sector) + sizeof(unsigned int)*num_block; 
    CU_ASSERT_EQUAL_FATAL(f_ctrl.blk_base, len);


    /* 
     * La root directory table deve contenere 3 entry valide
     * ".", ".." e "dir1" rispettivamente.
     * Prima mi sposto nel blocco giusto....
     */
    fseek(fs, f_ctrl.blk_base + (ROOT_IDX)*f_ctrl.b_sector.block_size, SEEK_SET); 

    /* ... e poi leggo l'entry "." */
    ret = fread(&root_info, sizeof(struct dir_entry), 1, fs); 
    CU_ASSERT_EQUAL_FATAL(ret, 1);
    CU_ASSERT_EQUAL_FATAL(root_info.used, DIR_ENTRY_BUSY);
    CU_ASSERT_STRING_EQUAL_FATAL(root_info.name, ".");
    CU_ASSERT_EQUAL_FATAL(root_info.attr, SUB_ENTRY);
    CU_ASSERT_EQUAL_FATAL(root_info.index, ROOT_IDX);
    /* .... l'entry ".."  */
    ret = fread(&root_info, sizeof(struct dir_entry), 1, fs);
    CU_ASSERT_EQUAL_FATAL(ret, 1);
    CU_ASSERT_EQUAL_FATAL(root_info.used, DIR_ENTRY_BUSY);
    CU_ASSERT_STRING_EQUAL_FATAL(root_info.name, "..");
    CU_ASSERT_EQUAL_FATAL(root_info.attr, SUB_ENTRY);
    CU_ASSERT_EQUAL_FATAL(root_info.index, ROOT_IDX);
    /* .. e l'entry "dir1" */
    ret = fread(&root_info, sizeof(struct dir_entry), 1, fs); 
    CU_ASSERT_EQUAL_FATAL(ret, 1);
    CU_ASSERT_EQUAL_FATAL(root_info.used, DIR_ENTRY_BUSY);
    CU_ASSERT_STRING_EQUAL_FATAL(root_info.name, "dir1");
    CU_ASSERT_EQUAL_FATAL(root_info.attr, SUB_ENTRY);
    index_dir1 = root_info.index;
    /* Le entry restanti devono essere vuote */
    b_size = f_ctrl.b_sector.block_size;
    n_entr = b_size / sizeof(struct dir_entry);
    for (i=3; i<n_entr; i++) {
	fread(&root_info, sizeof(struct dir_entry), 1, fs);
	CU_ASSERT_EQUAL_FATAL(root_info.used, DIR_ENTRY_FREE);
    }

    /*
     * Controlla che la directory table per /dir1
     * contenga due entry ("." e "..").
     */
    len = f_ctrl.blk_base + (index_dir1)*f_ctrl.b_sector.block_size;
    fseek(fs, len, SEEK_SET); 
    fread(&root_info, sizeof(struct dir_entry), 1, fs); /* Lettura "."  */
    CU_ASSERT_EQUAL_FATAL(root_info.used, DIR_ENTRY_BUSY);
    CU_ASSERT_STRING_EQUAL_FATAL(root_info.name, ".");
    CU_ASSERT_EQUAL_FATAL(root_info.attr, SUB_ENTRY);

    fread(&root_info, sizeof(struct dir_entry), 1, fs); /* Lettura ".."*/
    CU_ASSERT_EQUAL_FATAL(root_info.used, DIR_ENTRY_BUSY);
    CU_ASSERT_STRING_EQUAL_FATAL(root_info.name, "..");
    CU_ASSERT_EQUAL_FATAL(root_info.attr, SUB_ENTRY);
    CU_ASSERT_EQUAL_FATAL(root_info.index, ROOT_IDX);

    for (i=2; i<n_entr; i++) {
	fread(&root_info, sizeof(struct dir_entry), 1, fs);
	CU_ASSERT_EQUAL_FATAL(root_info.used, DIR_ENTRY_FREE);
    }

    /*
     * Controlliamo ora la FAT: 
     * esattamente due blocchi devono essere marcati LAST_BLOCK 
     * e gli altri BLOCK_FREE 
     */
    f_ptr = f_ctrl.fat_ptr;
    cnt = 0; 
    for (i=0; i<f_ctrl.b_sector.num_block; i++) {
	if (f_ptr[i] == LAST_BLOCK)
	    cnt++;
	else /* e controllo che gli altri siano free */
	    CU_ASSERT_EQUAL_FATAL(f_ptr[i], BLOCK_FREE);
    }
    CU_ASSERT_EQUAL_FATAL(cnt, 2);

    /*
     * Controlla la lunghezza dei file
     */
    len = sizeof(struct boot_sector);
    len = len + sizeof(unsigned int) * f_ctrl.b_sector.num_block;
    len = len + f_ctrl.b_sector.block_size * f_ctrl.b_sector.num_block;
    stat(fname, &st);
    CU_TEST_FATAL(len == st.st_size);

    /*
     * controlla il listing della directory appena creata
     */
    ret = fat_ls(fs, &f_ctrl, "/", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ex_list);

    /*
     * Cleanup
     */
    fclose(fs);
    unlink(fname); 
}


/**
 * \test
 *
 * Questo test ha lo scopo di controllare che venga gestita correttamente la
 * situazione in cui occorre aggiungere un nuovo blocco per contenere una
 * directory table. In particolare il test viene realizzato come segue:
 * 
 * - Viene richiesta la creazione di \e N directory sotto "/" in 
 *   modo da riempire il primo blocco della directory table.
 * - Viene creata una nuova directory in modo che la funzione 
 *   \c fat_mkdir aggiunga un nuovo blocco alla directory table.
 *
 * Il  test termina con successo se sono verificate le seguenti 
 * condizioni: 
 *
 * - La directory table deve essere distribuita su due blocchi.
 * - Il listing delle directory deve tornare tutte le subdirectory
 *   create.
 * 
 * \param fname il nome del file utilizzato per memorizzare il filesystem FAT.
 * \param block_size la dimensione di ciascun blocco delfilesystem FAT.
 * \param num_block numero di blocchi utilizzati dal filesystem FAT.
 */
void test_add_blk (char *fname, int block_size, unsigned int num_block)
{
    FILE *fs;
    struct fat_ctrl f_ctrl;
    unsigned int data_reg;
    int i, n_entr, b_size, len, ret;
    char dir_name[MAX_LEN_NAME + 2], *exp_list, *list;
     
    /*
     * Formatta il file ed effettua
     * il mounting
     */
    fs = fopen(fname, "w");
    fat_format(fs, block_size, num_block);
    fclose(fs);
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Crea alcune directory in modo da riempire il 
     * primo blocco della directory table
     */
    data_reg = f_ctrl.blk_base;                /* Inizio primo blocco */
    b_size = f_ctrl.b_sector.block_size;       /* Size del blocco */
    n_entr = b_size / sizeof(struct dir_entry); /* Numero di entry per blocco */

    
    for (i=0; i<n_entr - 2; i++) {
	snprintf(dir_name, MAX_LEN_NAME + 2, "/d%i", i);
	ret = fat_mkdir(fs, &f_ctrl, dir_name);
	CU_ASSERT_EQUAL_FATAL(ret, 0);
    }

    /*... aggiunge una nuova entry, cosi' che e' necessario un nuovo blocco */
    snprintf(dir_name, MAX_LEN_NAME + 2, "/d%i", n_entr - 2);
    ret = fat_mkdir(fs, &f_ctrl, dir_name);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Rimonta il file system per verificare che la struttura sia corretta
     */
    fclose(fs);
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /* 
     * Controlliamo che tutta la directory sia stata creata 
     * correttamente attraverso il listing 
     */
    ret = fat_ls(fs, &f_ctrl, "/", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /* Costruisce la stringa attesa  */
    exp_list = malloc(MAX_LEN_NAME + 1);
    CU_ASSERT_PTR_NOT_NULL_FATAL(exp_list);

    snprintf(exp_list,MAX_LEN_NAME + 1, "%s", ".\t..");
    for (i=0; i<n_entr - 1; i++) {
	snprintf(dir_name, MAX_LEN_NAME + 2,  "d%i", i);
	len = strlen(exp_list) + strlen(dir_name);
	exp_list = realloc(exp_list, len + 2);
	strcat(exp_list, "\t");
	strcat(exp_list, dir_name);
    }
    CU_ASSERT_STRING_EQUAL_FATAL(list, exp_list);
    free(list);

    for (i=0; i<n_entr - 1; i++) {
	snprintf(dir_name, MAX_LEN_NAME + 2,  "/d%i", i);
	ret = fat_ls(fs, &f_ctrl, dir_name, &list);
	CU_ASSERT_EQUAL_FATAL(ret, 0);
	CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..");
	free(list);
    }

    /*
     * Cleanup
     */
    fclose(fs);
    unlink(fname); 
}

/**
 * \test
 *
 * Questo test ha lo scopo di controllare che la creazione di\e nested 
 * directory venga gestita correttamente dalla funzione \c fat_mkdir.
 * In particolare questo test viene realizzato come segue:
 * 
 * - Viene generato l'albero di directory illustrato nella figura sotostante
 *   utilizzando la funzione \c fat_mkdir.
 * - Tramite la funzione \c fat_ls si controlla che ciascuna direcotry
 *   contenga il corretto insieme di subdirectory (incluso "." e "..").
 * - Si controlla che la duplicazione di una directory non a livello
 *   di root sia gestita correttamente.
 * - Si controlla che il tentativo di effettuare il listing di una directory 
 *   non esistente sia gestito correttamente.
 * 
 * \image html dirTest.jpg "Albero delle direcotry utilizato per il test" 
 * 
 * \param fname il nome del file utilizzato per memorizzare il filesystem FAT.
 * \param block_size la dimensione di ciascun blocco delfilesystem FAT.
 * \param num_block numero di blocchi utilizzati dal filesystem FAT.
 */
void test_recursive(char *fname, int block_size, unsigned int num_block)
{
    FILE *fs;
    struct fat_ctrl f_ctrl;
    int ret;
    char *list;

    /*
     * Formatta il filesystem
     */
    fs = fopen(fname, "w");
    fat_format(fs, block_size, num_block);
    fclose(fs);
    /* ... lo carica */
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Crea le directory D1, D2, D3 e D4 nella 
     * root directory ("/").
     */
    ret = fat_mkdir(fs, &f_ctrl, "/D1");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D2");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D3");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D4");
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Crea le directory D2_1, D2_2 e D2_3 nella
     * directory D2.
     */
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_1");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_2");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_3");
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Crea le directory D2_3_1 e D2_3_2 nella
     * directory D2_3.
     */
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_3/D2_3_1");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_3/D2_3_2");
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Crea D4_1 e D4_2 nella
     * directory D4.
     */
    ret = fat_mkdir(fs, &f_ctrl, "/D4/D4_1");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D4/D4_2");
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Controlla il contenuto della root
     */
    ret = fat_ls(fs, &f_ctrl, "/", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..\tD1\tD2\tD3\tD4");
    free(list);

    /*
     * Controlla il contenuto di /D1 
     */
    ret = fat_ls(fs, &f_ctrl, "/D1", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..");
    free(list);

    /*
     * Controlla il contenuto di /D2
     */
    ret = fat_ls(fs, &f_ctrl, "/D2", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..\tD2_1\tD2_2\tD2_3");
    free(list);
    
    /*
     * Controlla /D2/D2_1 e /D2/D2_2 
     */
    ret = fat_ls(fs, &f_ctrl, "/D2/D2_1", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..");
    free(list);
    ret = fat_ls(fs, &f_ctrl, "/D2/D2_2", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..");
    free(list);
    /*
     * Controlla /D2/D2_3 
     */
    ret = fat_ls(fs, &f_ctrl, "/D2/D2_3", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0); 
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..\tD2_3_1\tD2_3_2");

    /*
     * Controlla /D2/D2_3/D2_3_1 e /D2/D2_3/D2_3_2
     */
    ret = fat_ls(fs, &f_ctrl, "/D2/D2_3/D2_3_1", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0); 
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..");
    ret = fat_ls(fs, &f_ctrl, "/D2/D2_3/D2_3_2", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0); 
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..");
    
    /*
     * Controlla /D3
     */
    ret = fat_ls(fs, &f_ctrl, "/D3", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..");

    /*
     * Controlla /D4
     */
    ret = fat_ls(fs, &f_ctrl, "/D4", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..\tD4_1\tD4_2");

    /*
     * Controllo per directory duplicate
     */
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_3/D2_3_2");
    CU_ASSERT_EQUAL_FATAL(ret, EDAEX);
    ret = fat_ls(fs, &f_ctrl, "/D2/D4_2/D4_2_1", &list);
    CU_ASSERT_EQUAL_FATAL(ret, ENSFD);

    /*
     * Cleanup
     */
    fclose(fs);
    unlink(fname); 
}

/**
 * \test
 *
 * Questo test ha lo scopo di controllare la navigazione all'interno
 * del file system FAT utilizzando path che includiono "." e ".."
 * In particolare questo test viene realizzato come segue:
 * 
 * - Viene generato l'albero di directory utilizzato per il test 
 *   \c test_recursive_128.
 * - Tramite la funzione \c fat_ls si controlla che il contenuto delle
 *   seguenti directories sia corretto:
 *   - /D4/.
 *   - /D2/D2_3/..
 *   - /D2/D2_3/../.
 *   - /D4/D4_2/../../D2/D2_3
 *   - /D4/D4_2/../../D2/D2_3/D2_3_1/../../../D4
 * 
 * \param fname il nome del file utilizzato per memorizzare il filesystem FAT.
 * \param block_size la dimensione di ciascun blocco delfilesystem FAT.
 * \param num_block numero di blocchi utilizzati dal filesystem FAT.
 */
void test_rel_path(char *fname, int block_size, unsigned int num_block)
{
    FILE *fs;
    struct fat_ctrl f_ctrl;
    int ret;
    char *list;

    /*
     * Formatta il filesystem ....
     */
    fs = fopen(fname, "w");
    fat_format(fs, block_size, num_block);
    fclose(fs);

    /* lo carica ... */
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * D1, D2, D3 e D4 
     */
    ret = fat_mkdir(fs, &f_ctrl, "/D1");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D2");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D3");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D4");
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * D2_1, D2_2 e D2_3 
     */
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_1");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_2");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_3");
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * D2_3_1, D2_3_2 e D2_3.
     */
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_3/D2_3_1");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D2/D2_3/D2_3_2");
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * D4_1 e D4_2
     */
    ret = fat_mkdir(fs, &f_ctrl, "/D4/D4_1");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    ret = fat_mkdir(fs, &f_ctrl, "/D4/D4_2");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    
    /*
     * Controlla /D4/. 
     */
    ret = fat_ls(fs, &f_ctrl, "/D4/.", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..\tD4_1\tD4_2");

    /*
     * Controlla /D2/D2_3/.. 
     */
    ret = fat_ls(fs, &f_ctrl, "/D2/D2_3/..", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..\tD2_1\tD2_2\tD2_3");

    /*
     * Controlla /D2/D2_3/../. 
     */
    ret = fat_ls(fs, &f_ctrl, "/D2/D2_3/../.", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..\tD2_1\tD2_2\tD2_3");


    /*
     * Controlla /D4/D4_2/../../D2/D2_3 
     */
    ret = fat_ls(fs, &f_ctrl, "/D4/D4_2/../../D2/D2_3", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0); 
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..\tD2_3_1\tD2_3_2");

    /*
     * Controlla /D4/D4_2/../../D2/D2_3/D2_3_1/../../../D4 
     */
    ret = fat_ls(fs, &f_ctrl, "/D4/.", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ".\t..\tD4_1\tD4_2");

    /*
     * Cleanup
     */
    fclose(fs);
    unlink(fname); 
}


/**
 * Test creazone di una directory per un filsystem formattato
 * utilizzando 10 blocchi di dimensione 128 Bytes ciascuno.
 */
void test_mkdir_128 (void)
{
    run_mkdir("DataFile/fat_128", 128, 10);
}

/**
 * Test creazone di una directory per un filsystem formattato
 * utilizzando 10 blocchi di dimensione 1KB ciascuno.
 */
void test_mkdir_1K (void)
{
    run_mkdir("DataFile/fat_1K", 1024, 10);
}

/**
 * Test creazone di una directory per un filsystem formattato
 * utilizzando 10 blocchi di dimensione 2KB ciascuno.
 */
void test_mkdir_2K (void)
{
    run_mkdir("DataFile/fat_2K", 1024*2, 10);
}

/**
 * Test creazone di una directory per un filsystem formattato
 * utilizzando 10 blocchi di dimensione 4KB ciascuno.
 */
void test_mkdir_4K (void)
{
    run_mkdir("DataFile/fat_4K", 1024*4, 10);
}

/**
 * Test di aggiunta di un blocco alla directory table per un filsystem
 *  formattato utilizzando 10 blocchi di dimensione 128 Bytes ciascuno.
 */
void test_add_blk_128 (void)
{
    test_add_blk("DataFile/fat_128", 128, 256);
}

/**
 * Test di aggiunta di un blocco alla directory table per un filsystem
 *  formattato utilizzando 2K blocchi di dimensione 1KB ciascuno.
 */
void test_add_blk_1K (void)
{
    test_add_blk("DataFile/fat_1K", 1024, 1024*2);
}

/**
 * Test di aggiunta di un blocco alla directory table per un filsystem
 *  formattato utilizzando 4K blocchi di dimensione 1KB ciascuno.
 */
void test_add_blk_2K (void)
{
    test_add_blk("DataFile/fat_2K", 1024*2, 1024*4);
}

/**
 * Test di aggiunta di un blocco alla directory table per un filsystem
 *  formattato utilizzando 8K blocchi di dimensione 1KB ciascuno.
 */
void test_add_blk_4K (void)
{
    test_add_blk("DataFile/fat_4K", 1024*4, 1024*8);
}

/**
 * Test creazone di un albero di directories per un filsystem formattato
 * utilizzando 20 blocchi di dimensione 128 Bytes ciascuno.
 */
void test_recursive_128 (void)
{
    test_recursive("DataFile/fat_128", 128, 20);
}

/**
 * Test creazone di un albero di directories per un filsystem formattato
 * utilizzando 20 blocchi di dimensione 1KB ciascuno.
 */
void test_recursive_1K (void)
{
    test_recursive("DataFile/fat_1K", 1024, 20);
}

/**
 * Test creazone di un albero di directories per un filsystem formattato
 * utilizzando 20 blocchi di dimensione 2KB ciascuno.
 */
void test_recursive_2K (void)
{
    test_recursive("DataFile/fat_2K", 1024*2, 20);
}

/**
 * Test creazone di un albero di directories per un filsystem formattato
 * utilizzando 20 blocchi di dimensione 4KB ciascuno.
 */
void test_recursive_4K (void)
{
    test_recursive("DataFile/fat_4K", 1024*4, 20);
}

/**
 * Test navigazione tramie path relativi  per un filsystem formattato
 * utilizzando 20 blocchi di dimensione 128 ciascuno.
 */
void test_rel_path_128 (void)
{
    test_rel_path("DataFile/fat_128", 128, 20);
}

/**
 * Test navigazione tramie path relativi  per un filsystem formattato
 * utilizzando 20 blocchi di dimensione 1KB ciascuno.
 */
void test_rel_path_1K (void)
{
    test_rel_path("DataFile/fat_1K", 1024, 20);
}

/**
 * Test navigazione tramie path relativi  per un filsystem formattato
 * utilizzando 20 blocchi di dimensione 2KB ciascuno.
 */
void test_rel_path_2K (void)
{
    test_rel_path("DataFile/fat_2K", 1024*2, 20);
}

/**
 * Test navigazione tramie path relativi  per un filsystem formattato
 * utilizzando 20 blocchi di dimensione 4KB ciascuno.
 */
void test_rel_path_4K (void)
{
    test_rel_path("DataFile/fat_4K", 1024*4, 20);
}

/**
 * Test suite per la funzione  \c fat_mkdir
 */
void add_suite_mkdir(void)
{
    CU_pSuite mkdir_suite = NULL;

    mkdir_suite = CU_add_suite("Creating Directory", mkdir_init, mkdir_clean);
    /* Bad path test. */
    CU_add_test(mkdir_suite, "Incorrect pathname detection", test_pathname); 
    /* Simple directory creation tests. */ 
    CU_add_test(mkdir_suite, "Make a directory with block size 128 bytes", test_mkdir_128); 
    CU_add_test(mkdir_suite, "Make a directory with block size 1KB", test_mkdir_1K); 
    CU_add_test(mkdir_suite, "Make a directory with block size 2KB", test_mkdir_2K);
    CU_add_test(mkdir_suite, "Make a directory with block size 4KB", test_mkdir_4K);  
    /* Add directiry table block tests. */
    CU_add_test(mkdir_suite, "Exceeding block size for directory table (128 bytes)", test_add_blk_128);
    CU_add_test(mkdir_suite, "Exceeding block size for directory table (1KB)", test_add_blk_1K);
    CU_add_test(mkdir_suite, "Exceeding block size for directory table (2KB)", test_add_blk_2K);
    CU_add_test(mkdir_suite, "Exceeding block size for directory table (4KB)", test_add_blk_4K);
    /* Nested directory tests. */
    CU_add_test(mkdir_suite, "Nested directories creation (128 bytes)", test_recursive_128);
    CU_add_test(mkdir_suite, "Nested directories creation (1KB)", test_recursive_1K);
    CU_add_test(mkdir_suite, "Nested directories creation (2KB)", test_recursive_2K);
    CU_add_test(mkdir_suite, "Nested directories creation (4KB)", test_recursive_4K);
    /* Relative path navigation tests. */
    CU_add_test(mkdir_suite, "Relative path navigation (128 bytes)", test_rel_path_128 );
    CU_add_test(mkdir_suite, "Relative path navigation (1KB)", test_rel_path_1K );
    CU_add_test(mkdir_suite, "Relative path navigation (2KB)", test_rel_path_2K );
    CU_add_test(mkdir_suite, "Relative path navigation (4KB bytes)", test_rel_path_4K );
}
