#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <mcheck.h>
#include "CUnit.h"
#include "TestDB.h"
#include "Basic.h"
#include "fat.h"
#include "format.h"
#include "load_fat.h"
#include "list.h"
#include "open_fat.h"
#include "write_fat.h"
#include "read_fat.h"
#include "rw_suite.h"
#define N 25
/**
 * \file 
 * Questo file definisce la test suite per le funzioni di scrittura e
 * lettura di un file all'interno di un filesystem SFAT.
 */

/** Funzione di inizializzazione della suite  */
int rw_init(void) { return 0; }

/** Funzione di inizializzazione della suite  */
int rw_clean(void) { return 0; }

/**
 * \test
 *
 * Questa funzione effettua un test basico della scrittura e lettura di un
 * file. In particolare la funzione effettua le seguenti operazioni:
 *
 * - Crea il file \c rw_char nella root directory.
 * - Effettua la scrittura di N caratteri (da a in poi) un carattere per volta.
 * - Controlla che la lunghezza del file sia cambiata da 0 a N.
 * - Controlla che il listing della root directory sia corretto.
 * - Richiede la lettura di ciascun carattere con operazioni di read separate
 *   e controlla che il risultato sia corretto.
 * - Richiede la lettura dei N caratteri in un'unica sequenza di byte e 
 *   controlla che il risultato sia corretto.
 *
 * Inoltre la funzione crea un secondo file \c rw_int ed effettua la scrittura
 * di N interi (da 0 a N-1). Su tale file vengono effettuati gli stessi tipi di
 * controlli previsti per il file \c rw_char.
 */
void basic_rw (char *fname, int block_size, unsigned int num_block)
{
 FILE *fs;
    struct fat_ctrl f_ctrl;
    struct dir_entry root_info;
    unsigned int len;
    int  ret, i, int_buff_in[N], int_buff_out[N];
    char *list, buff_in[N], buff_out[N], c;
    char ex_list1[ ] = ".\t..\trw_char";

    /*
     * Formatta il filesystem e ...
     */
    fs = fopen(fname, "w");
    fat_format(fs, block_size, num_block);
    fclose(fs);
    /* ... lo carica */
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Crea il file rw_char e chiude il fs ...
     */
    ret = fat_open(fs, &f_ctrl, "/rw_char");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    fclose(fs);
    /* ... cerca di montare il fs */
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Prepara il buffer ...
     */

    c = 'a'; /* inizia dal carattere 'a' */
    for (i=0; i<N; i++)
	buff_in[i] = c + i;

    /* ... scrive il buffer su file e chiude il file system. */
    ret = fat_write(fs, &f_ctrl, "/rw_char", buff_in, N*sizeof(char));
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    fclose(fs);

    /*
     * Effettua di nuovo il mounting del file system 
     */
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    
    /*
     * Controlla la lunghezza del file
     */
    len = sizeof(struct dir_entry)*2;
    len = f_ctrl.blk_base + (ROOT_IDX)*f_ctrl.b_sector.block_size+len;
    fseek(fs, len ,SEEK_SET); 
    ret = fread(&root_info, sizeof(struct dir_entry), 1, fs); 
    CU_ASSERT_EQUAL_FATAL(ret, 1);
    CU_ASSERT_EQUAL_FATAL(root_info.len, N*sizeof(char));
    
    /*
     * Controlla il listing di "/"
     */
    ret = fat_ls(fs, &f_ctrl, "/", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ex_list1);

    /*
     * Effettua la lettura di controllo
     */
    c = 'a'; 

    for (i=0; i<N; i++) {
	ret = fat_read(fs, &f_ctrl, "/rw_char", i, sizeof(char), buff_out);
	CU_ASSERT_EQUAL_FATAL(ret, sizeof(char));
	CU_ASSERT_FATAL(buff_out[0] == c);
	c = c + 1;
    }

    /*
     * Legge N byte dal file e controlla il risultato
     */
    ret = fat_read(fs, &f_ctrl, "/rw_char", 0, sizeof(char)*N, buff_out);
    CU_ASSERT_EQUAL_FATAL(ret, sizeof(char)*N);
    for (i=0; i<N; i++) 
	CU_ASSERT_FATAL(buff_out[i] == buff_in[i]);


    /*
     * Stesso test su N interi (da 0 a N-1)
     */
    ret = fat_open(fs, &f_ctrl, "/rw_int");

    /* Inizializza il buffer*/
    for (i=0; i<N; i++)
	int_buff_in[i] = i;

    /* Scrive il buffer sul file rw_int */
    ret = fat_write(fs, &f_ctrl, "/rw_int", (char *)int_buff_in, N*sizeof(int));
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    fclose(fs);

    /* Mounting del filesystem */
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /* Check della lunghezza del file */
    len = sizeof(struct dir_entry)*3;
    len = f_ctrl.blk_base + ROOT_IDX*f_ctrl.b_sector.block_size + len;
    fseek(fs, len ,SEEK_SET); 
    ret = fread(&root_info, sizeof(struct dir_entry), 1, fs); 
    CU_ASSERT_EQUAL_FATAL(ret, 1);
    CU_ASSERT_EQUAL_FATAL(root_info.len, N*sizeof(int));

    /* Controlla la lettura di un intero alla volta */
    for (i=0; i<N; i++) {
	ret = fat_read(fs, &f_ctrl, "/rw_int", i*sizeof(int), sizeof(int), (char *)int_buff_out);
	CU_ASSERT_EQUAL_FATAL(ret, sizeof(int));
	CU_ASSERT_FATAL((int)int_buff_out[0] == i);
    }
    
    /*  Legge tutti gli interi dal file. */
    ret = fat_read(fs, &f_ctrl, "/rw_int", 0, sizeof(int)*N, (char *)int_buff_out);
    CU_ASSERT_EQUAL_FATAL(ret, sizeof(int)*N);
    for (i=0; i<N; i++) 
	CU_ASSERT_FATAL(int_buff_out[i] == int_buff_in[i]);

    fclose(fs);

}

/**
 * \test
 *
 * Questa funzione effettua dei test di alcune delle condizioni limite
 * delle funzione di scrittura e
 * lettura di un file all'interno di un filesystem SFAT. Per questo test
 * verra' utilizzata una struttura che contiene un intero ed un carattere.
 * I test che vengono realizzati da
 * questa funzione sono i seguenti:
 *
 * - Vengono effettuate una serie di scritture in modo da riempire un blocco.
 *   Vengono poi effettuate una serie di letture per assicurarsi che le scritture
 *   siano avvenute correttamente alla fine del file.
 * - Viene effettuata una nuova scrittura che richiede l'allocazione di un nuovo 
 *   blocco. Viene eseguita una lettura per verificare la correttezza della precedente scrittura.
 * - Viene effettuata una lettura della dimensione di due blocchi in modo da 
 *   verificare che venga rilevata correttamente la fine di un file.
 * - Viene effettuata una serie di scritture in modo da utilizzare tutti i blocchi 
 *   disponibili all'interno del filesystem (\c num_block).
 * - Vengono effettuate delle letture random che accedono a varie parti del file.
 *   Lo scopo e' quello di verificare che la lettura di porzioni del file avvenga
 *   correttamente. 
 *
 */
void advanced_rw (char *fname, int block_size, unsigned int num_block)
{
    /* struttura da scrivere nel file */
    struct test_data {
	char c_data;
	int  i_data;
    };
    FILE *fs;
    struct fat_ctrl f_ctrl;
    int  n_blk, r_bytes, ret, i, num_elem, begin;
    char c;
    struct test_data *buff_in, *buff_out;

    /*
     * Formattazione del fs e ...
     */
    fs = fopen(fname, "w");
    fat_format(fs, block_size, num_block);
    fclose(fs);
    /* ... caricamento */
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Crate the file rw_struct and ...
     */
    ret = fat_open(fs, &f_ctrl, "/rw_st");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    fclose(fs);
    /* ... try to mount the filsystem */
    fs = fopen(fname, "r+");
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Numero di scritture necessarie e riempire il blocco 
     * 
     */
    num_elem = block_size / sizeof(struct test_data);
    buff_in = (struct test_data *)malloc(sizeof(struct test_data)*num_elem);

    c = 0x00; /* Starts from '\0' */
    for (i=0; i<num_elem; i++) {
	buff_in[i].c_data = (c + i) % 128;
	buff_in[i].i_data = i;
	ret = fat_write(fs, &f_ctrl, "/rw_st", (char *)(buff_in + i), sizeof(struct test_data));
	CU_ASSERT_EQUAL_FATAL(ret, 0);
    }

    /*
     * Lettura e controllo delle strutture scritte
     */
    buff_out = (struct test_data *)malloc(sizeof(struct test_data)*num_elem );
    ret = fat_read(fs, &f_ctrl, "/rw_st", 0, sizeof(struct test_data)*num_elem, (char *)buff_out);
    CU_ASSERT_EQUAL_FATAL(ret, sizeof(struct test_data)*num_elem);
    for (i=0; i<num_elem; i++) { 
	CU_ASSERT_FATAL(buff_out[i].c_data == buff_in[i].c_data);
	CU_ASSERT_FATAL(buff_out[i].i_data == buff_in[i].i_data);
    }

    /*
     * Scriviamo una nuova struttua, 
     * questo richiede l'allocazione di un nuovo blocco ..
     */
    buff_in[0].c_data = num_elem % 128;
    buff_in[0].i_data = num_elem;
    ret = fat_write(fs, &f_ctrl, "/rw_st", (char *)(buff_in), sizeof(struct test_data));
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*... rilettura .... */
    ret = fat_read(fs, &f_ctrl, "/rw_st", num_elem*sizeof(struct test_data),
		   sizeof(struct test_data), (char *)buff_out);
    CU_ASSERT_EQUAL_FATAL(ret, sizeof(struct test_data));
    CU_ASSERT_FATAL(buff_out[0].c_data == buff_in[0].c_data);
    CU_ASSERT_FATAL(buff_out[0].i_data == buff_in[0].i_data);

    free(buff_in);
    free(buff_out);
    /*
     * Leggiamo 2*num_elem strutture da file. Solo num_elem + 1 devono essere
     * ritornate nel buffer   
     */
    r_bytes = sizeof(struct test_data)*2*num_elem;
    buff_out = (struct test_data *)malloc(r_bytes);
    CU_ASSERT_FATAL(buff_out!=NULL);

    ret = fat_read(fs, &f_ctrl, "/rw_st", 0, r_bytes, (char *)buff_out);
    CU_ASSERT_EQUAL_FATAL(ret, sizeof(struct test_data)*(num_elem + 1));
    for (i=0; i<num_elem + 1; i++) { 
	CU_ASSERT_FATAL(buff_out[i].i_data == i);
	CU_ASSERT_FATAL(buff_out[i].c_data == i%128);
    }

    /*
     * Riempiamo i blocchi disponibili. 
     * Nota: un blocco e' usato dalla root e due dal file rw_st.
     */
     n_blk = num_block - 2;
     for (i=0; i<(n_blk*num_elem)-1; i++) {
	 buff_in[0].c_data = (num_elem + 1 + i) % 128;
	 buff_in[0].i_data = num_elem + 1 + i;
	 ret = fat_write(fs, &f_ctrl, "/rw_st", (char *)(buff_in), sizeof(struct test_data));
	 CU_ASSERT_EQUAL_FATAL(ret, 0);
     }

     /* Reads all data */
     free(buff_out);
     r_bytes = sizeof(struct test_data)*num_elem*(num_block-1);
     buff_out = (struct test_data *)malloc(r_bytes); 
     ret = fat_read(fs, &f_ctrl, "/rw_st", 0, r_bytes, (char *)buff_out);
     CU_ASSERT_EQUAL_FATAL(ret, r_bytes);
     for (i=0; i<num_elem*(num_block-1); i++) { 
	 CU_ASSERT_FATAL(buff_out[i].c_data == i%128);
	 CU_ASSERT_FATAL(buff_out[i].i_data == i);
     }
     
    /* Legge dalla struttura 4 nel blocco 2 alla 3 nel blocco 4 . */
    begin = (num_elem*2 + 4)*sizeof(struct test_data);
    r_bytes = ((sizeof(struct test_data)*num_elem*3) + 3*sizeof(struct test_data) - begin);
    ret = fat_read(fs, &f_ctrl, "/rw_st", begin, r_bytes, (char *)buff_out);
    CU_ASSERT_EQUAL_FATAL(ret, r_bytes);
     for (i=num_elem*2 + 4; i<num_elem*3 + 3; i++) { 
	 CU_ASSERT_FATAL(buff_out[i].c_data == i%128);
	 CU_ASSERT_FATAL(buff_out[i].i_data == i);
     }

     /* Legge i dati dell'ultimo blocco*/
     begin = (num_elem*(num_block - 2)) * sizeof(struct test_data);
     r_bytes = sizeof(struct test_data)*num_elem;
     ret = fat_read(fs, &f_ctrl, "/rw_st", begin, r_bytes, (char *)buff_out);
     CU_ASSERT_EQUAL_FATAL(ret, r_bytes);
     for (i=num_elem*(num_block - 2); i<num_elem*(num_block - 1); i++) { 
	 CU_ASSERT_FATAL(buff_out[i].c_data == i%128);
	 CU_ASSERT_FATAL(buff_out[i].i_data == i);
     }

     /*
      * Check file system full.
      */
     ret = fat_write(fs, &f_ctrl, "/rw_st", (char *)(buff_in), sizeof(struct test_data));
     CU_ASSERT_EQUAL_FATAL(ret, ENMSD);
}

/**
 * Test di base per la lettura e scrittura di un file per un filsystem
 * formattato utilizzando 10 blocchi di dimensione 128 Bytes ciascuno.
 */
void basic_rw_128 (void)
{
    basic_rw("DataFile/fat_128", 128, 10);
}

/**
 * Test di base per la lettura e scrittura di un file per un filsystem
 * formattato utilizzando 10 blocchi di dimensione 1KB ciascuno.
 */
void basic_rw_1K (void)
{
    basic_rw("DataFile/fat_1K", 1024, 10);
}

/**
 * Test di base per la lettura e scrittura di un file per un filsystem
 * formattato utilizzando 10 blocchi di dimensione 2KB ciascuno.
 */
void basic_rw_2K (void)
{
    basic_rw("DataFile/fat_2K", 1024*2, 10);
}

/**
 * Test di base per la lettura e scrittura di un file per un filsystem
 * formattato utilizzando 10 blocchi di dimensione 4KB ciascuno.
 */
void basic_rw_4K (void)
{
    basic_rw("DataFile/fat_4K", 1024*4, 10);
}

/**
 * Test avanzato per la lettura e scrittura di un file per un filsystem
 * formattato utilizzando 10 blocchi di dimensione 128 Bytes ciascuno.
 */
void advanced_rw_128 (void)
{
    advanced_rw("DataFile/fat_128", 128, 10);
}

/**
 * Test avanzato per la lettura e scrittura di un file per un filsystem
 * formattato utilizzando 10 blocchi di dimensione 1KB  ciascuno.
 */
void advanced_rw_1K (void)
{
    advanced_rw("DataFile/fat_1K", 1024, 10);
}

/**
 * Test avanzato per la lettura e scrittura di un file per un filsystem
 * formattato utilizzando 10 blocchi di dimensione 2KB  ciascuno.
 */
void advanced_rw_2K (void)
{
    advanced_rw("DataFile/fat_2K", 1024*2, 10);
}

/**
 * Test avanzato per la lettura e scrittura di un file per un filsystem
 * formattato utilizzando 10 blocchi di dimensione 4KB  ciascuno.
 */
void advanced_rw_4K (void)
{
    advanced_rw("DataFile/fat_4K", 1024*4, 10);
}


/**
 * Test suite per le funzioni  \c fat_read e \c fat_write
 */
void add_suite_rw(void)
{
    CU_pSuite rw_suite = NULL;

    rw_suite = CU_add_suite("File R/W operations", rw_init, rw_clean);

    /* Basic R/W operations test */ 
    CU_add_test(rw_suite, "Basic R/W operations with block size 128 bytes", basic_rw_128); 
    CU_add_test(rw_suite, "Basic R/W operations with block size 1K bytes", basic_rw_1K); 
    CU_add_test(rw_suite, "Basic R/W operations with block size 2K bytes", basic_rw_2K); 
    CU_add_test(rw_suite, "Basic R/W operations with block size 4K bytes", basic_rw_4K); 

    /* Advanced R/W operations test */ 
    CU_add_test(rw_suite, "Advanced R/W operations with block size 128 bytes", advanced_rw_128);
    CU_add_test(rw_suite, "Advanced R/W operations with block size 1K bytes", advanced_rw_1K);
    CU_add_test(rw_suite, "Advanced R/W operations with block size 2K bytes", advanced_rw_2K);
    CU_add_test(rw_suite, "Advanced R/W operations with block size 4K bytes", advanced_rw_4K);
}
