#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <mcheck.h>
#include "CUnit.h"
#include "TestDB.h"
#include "Basic.h"
#include "fat.h"
#include "format.h"
#include "load_fat.h"
#include "list.h"
#include "open_fat.h"
#include "open_suite.h"

/**
 * \file 
 * Questo file definisce la test suite per la funzione di creazione di
 * un file \c fat_open.
 */

/** Funzione di inizializzazione della suite  */
int open_init(void) {  return 0; }

/** Funzione di inizializzazione della suite  */
int open_clean(void) {    return 0; }

/**
 * \test
 *
 * Esegue il test di creazione di un file all'interno della root
 * directory. In particolare si controlla che:
 *
 * - La root directory contenga tre entry, due relative a "." e ".."
 *   ed una relativa al nuovo file.
 * - La lunghezza del file che contiene il filesystem SFAT sia corretta.
 * - Che il listing della root directory contenga le corrette informazioni.
 *
 */
void simple_open (char *fname, int block_size, unsigned int num_block)
{
    FILE *fs;
    struct fat_ctrl f_ctrl;
    struct dir_entry root_info;
    unsigned int len;
    int  ret;
    struct stat st;
    char *list;
    char ex_list[ ] = ".\t..\tfile1";

    /*
     * Formattazione file
     */
    if ( ( fs = fopen(fname, "w") ) == NULL ) {
	perror("simpleopen");
	exit (EXIT_FAILURE);
    }
    fat_format(fs, block_size, num_block);
    fclose(fs);
    /* caricamento  */
    if ( ( fs = fopen(fname, "r+") ) == NULL) {
	perror("simpleopen");
	exit (EXIT_FAILURE);
    }
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /*
     * Cerca di creare il file "file1" e chiude il file
     */
    ret = fat_open(fs, &f_ctrl, "/file1");
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    fclose(fs);
    /* cerca di fare il mount del filesystem .*/
    if ( ( fs = fopen(fname, "r+") ) == NULL) {
	perror("simpleopen");
	exit (EXIT_FAILURE);
    }
    ret = mount_fat(fs, &f_ctrl);
    CU_ASSERT_EQUAL_FATAL(ret, 0);

    /* 
     * La root directory table deve contenere tre entry
     * per ".", ".." e "dir1" rispettivamente.
     */
    fseek(fs, f_ctrl.blk_base + (ROOT_IDX)*f_ctrl.b_sector.block_size, SEEK_SET);
    /* Legge  "." entry */
    ret = fread(&root_info, sizeof(struct dir_entry), 1, fs); 
    CU_ASSERT_EQUAL_FATAL(ret, 1);
    CU_ASSERT_EQUAL_FATAL(root_info.used, DIR_ENTRY_BUSY);
    CU_ASSERT_STRING_EQUAL_FATAL(root_info.name, ".");
    CU_ASSERT_EQUAL_FATAL(root_info.attr, SUB_ENTRY);
    /* Legge  ".." entry */
    ret = fread(&root_info, sizeof(struct dir_entry), 1, fs);
    CU_ASSERT_EQUAL_FATAL(ret, 1);
    CU_ASSERT_EQUAL_FATAL(root_info.used, DIR_ENTRY_BUSY);
    CU_ASSERT_STRING_EQUAL_FATAL(root_info.name, "..");
    CU_ASSERT_EQUAL_FATAL(root_info.attr, SUB_ENTRY);
    /* Legge "file1" entry */
    ret = fread(&root_info, sizeof(struct dir_entry), 1, fs); 
    CU_ASSERT_EQUAL_FATAL(ret, 1);
    CU_ASSERT_EQUAL_FATAL(root_info.used, DIR_ENTRY_BUSY);
    CU_ASSERT_STRING_EQUAL_FATAL(root_info.name, "file1");
    CU_ASSERT_EQUAL_FATAL(root_info.attr, FILE_ENTRY);


    /*
     * Controlla la lunghezza del file
     */
    len = sizeof(struct boot_sector);
    len = len + sizeof(unsigned int) * f_ctrl.b_sector.num_block;
    len = len + f_ctrl.b_sector.block_size * f_ctrl.b_sector.num_block;
    stat(fname, &st);
    CU_TEST_FATAL(len == st.st_size);

    /*
     * Proviamo a listare il nuovo file.
     */
    ret = fat_ls(fs, &f_ctrl, "/", &list);
    CU_ASSERT_EQUAL_FATAL(ret, 0);
    CU_ASSERT_STRING_EQUAL_FATAL(list, ex_list);

    /*
     * Cleanup
     */
    fclose(fs);
    unlink(fname); 
}

/**
 * Test di base per la creazione di un file per un filsystem formattato
 * utilizzando 10 blocchi di dimensione 128 Byte ciascuno.
 */
void simple_open_128 (void)
{
    simple_open("DataFile/fat_128", 128, 10);
}

/**
 * Test di base per la creazione di un file per un filsystem formattato
 * utilizzando 10 blocchi di dimensione 1KB ciascuno.
 */
void simple_open_1K (void)
{
    simple_open("DataFile/fat_1K", 1024, 10);
}

/**
 * Test di base per la creazione di un file per un filsystem formattato
 * utilizzando 10 blocchi di dimensione 2KB ciascuno.
 */
void simple_open_2K (void)
{
    simple_open("DataFile/fat_2K", 1024*2, 10);
}

/**
 * Test di base per la creazione di un file per un filsystem formattato
 * utilizzando 10 blocchi di dimensione 4KB ciascuno.
 */
void simple_open_4K (void)
{
    simple_open("DataFile/fat_4K", 1024*4, 10);
}

/**
 * Test suite per la funzione  \c fat_open
 */
void add_suite_open(void)
{
    CU_pSuite open_suite = NULL;

    open_suite = CU_add_suite("File creation", open_init, open_clean);

    /* Test creazione file. */ 
    CU_add_test(open_suite, "Create a file with block size 128 bytes", simple_open_128); 
    CU_add_test(open_suite, "Create a file with block size 1KB", simple_open_1K); 
    CU_add_test(open_suite, "Create a file with block size 2KB", simple_open_2K);
    CU_add_test(open_suite, "Create a file with block size 4KB", simple_open_4K);  
}
